@@ -142,8 +142,8 @@ translate_message2(error(ISO, SWI)) -->
     swi_location(SWI),
     term_message(ISO),
     swi_extra(SWI).
-translate_message2('$aborted') -->
-    [ 'Execution Aborted' ].
+translate_message2(unwind(Term)) -->
+    unwind_message(Term).
 translate_message2(message_lines(Lines), L, T) :- % deal with old C-warning()
     make_message_lines(Lines, L, T).
 translate_message2(format(Fmt, Args)) -->
@@ -384,6 +384,11 @@ syntax_error(undefined_char_escape(C)) -->
     [ 'Unknown character escape in quoted atom or string: `\\~w\''-[C] ].
 syntax_error(void_not_allowed) -->
     [ 'Empty argument list "()"' ].
+syntax_error(Term) -->
+    { compound(Term),
+      compound_name_arguments(Term, Syntax, [Text])
+    }, !,
+    [ '~w expected, found '-[Syntax], ansi(code, '"~w"', [Text]) ].
 syntax_error(Message) -->
     [ '~w'-[Message] ].
 
@@ -637,6 +642,23 @@ thread_context -->
 thread_context -->
     [].
 
+		 /*******************************
+		 *        UNWIND MESSAGES	*
+		 *******************************/
+
+unwind_message(Var) -->
+    { var(Var) }, !,
+    [ 'Unknown unwind message: ~p'-[Var] ].
+unwind_message(abort) -->
+    [ 'Execution Aborted' ].
+unwind_message(halt(_)) -->
+    [].
+unwind_message(thread_exit(Term)) -->
+    [ 'Invalid thread_exit/1.  Payload: ~p'-[Term] ].
+unwind_message(Term) -->
+    [ 'Unknown "unwind" exception: ~p'-[Term] ].
+
+
                  /*******************************
                  *        NORMAL MESSAGES       *
                  *******************************/
@@ -1824,6 +1846,13 @@ deprecated(autoload(TargetModule, File, _M:PI, expansion)) -->
     load_file(File), [ ' into ' ],
     target_module(TargetModule),
     [ ' is deprecated due to term- or goal-expansion' ].
+deprecated(source_search_working_directory(File, _FullFile)) -->
+    [ 'Found file ', ansi(code, '~w', [File]),
+      ' relative to the current working directory.', nl,
+      'This behaviour is deprecated but still supported by', nl,
+      'the Prolog flag ',
+      ansi(code, source_search_working_directory, []), '.', nl
+    ].
 
 load_file(File) -->
     { file_base_name(File, Base),
