@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2006-2023, University of Amsterdam
+    Copyright (c)  2006-2024, University of Amsterdam
 			      VU University Amsterdam
 			      CWI, Amsterdam
 			      SWI-Prolog Solutions b.v.
@@ -81,9 +81,8 @@ please visit https://www.swi-prolog.org/pldoc/package/plunit.
 		 *    CONDITIONAL COMPILATION   *
 		 *******************************/
 
-swi     :- catch(current_prolog_flag(dialect, swi), _, fail), !.
-swi     :- catch(current_prolog_flag(dialect, yap), _, fail).
-sicstus :- catch(current_prolog_flag(system_type, _), _, fail).
+swi     :- catch(current_prolog_flag(dialect, swi),     _, fail).
+sicstus :- catch(current_prolog_flag(dialect, sicstus), _, fail).
 
 throw_error(Error_term,Impldef) :-
     throw(error(Error_term,context(Impldef,_))).
@@ -134,15 +133,25 @@ goal_expansion(current_module(Module,File),
 :- initialization init_flags.
 
 init_flags :-
-    (   global_test_option(Option, _Value, _Type, Default),
+    (   global_test_option(Option, _Value, Type, Default),
 	Default \== (-),
 	Option =.. [Name,_],
 	atom_concat(plunit_, Name, Flag),
-	create_prolog_flag(Flag, Default, [keep(true)]),
+        flag_type(Type, FlagType),
+	create_prolog_flag(Flag, Default, [type(FlagType), keep(true)]),
 	fail
     ;   true
     ).
 
+flag_type(boolean, FlagType) => FlagType = boolean.
+flag_type(Type, FlagType), Type = oneof(Atoms), maplist(atom, Atoms) =>
+    FlagType = Type.
+flag_type(oneof(_), FlagType) => FlagType = term.
+flag_type(positive_integer, FlagType) => FlagType = integer.
+flag_type(number, FlagType) => FlagType = float.
+
+
+
 %!  set_test_options(+Options)
 %
 %   Specifies how to deal with test suites.  Defined options are:
@@ -729,19 +738,23 @@ run_tests_sync(Units0, Options) :-
     asserta(test_count(Count)),
     save_output_state,
     setup_call_cleanup(
-	setup_jobs(Count),
-	setup_call_cleanup(
-	    setup_trap_assertions(Ref),
-	    ( call_time(run_units(Units, Options), Time),
-              test_summary(_All, Summary)
-            ),
-	    report_and_cleanup(Ref, Time, Options)),
-	cleanup_jobs),
+        setup_trap_assertions(Ref),
+        call_time(setup_jobs_and_run_units(Count, Units, Summary, Options),
+                  Time),
+        report_and_cleanup(Ref, Time, Options)),
     (   option(summary(Summary), Options)
     ->  true
     ;   test_summary_passed(Summary) % fail if some test failed
     ).
 
+setup_jobs_and_run_units(Count, Units, Summary, Options) :-
+    setup_call_cleanup(
+        setup_jobs(Count),
+        ( run_units(Units, Options),
+          test_summary(_All, Summary)
+        ),
+        cleanup_jobs).
+
 %!  report_and_cleanup(+Ref, +Time, +Options)
 %
 %   Undo changes to the environment   (trapping  assertions), report the
@@ -1010,12 +1023,12 @@ run_test(Unit, Name, Line, UnitOptions, Options, Body) :-
     option(forall(Generator), Options),
     !,
     unit_module(Unit, Module),
-    term_variables(Generator, Vars),
     start_test(Unit, @(Name,Line), Nth),
     State = state(0),
     call_time(forall(Module:Generator,            % may become concurrent
                      (   incr_forall(State, I),
-                         run_test_once6(Unit, Name, forall(Vars, Nth-I), Line,
+                         run_test_once6(Unit, Name,
+                                        forall(Generator, Nth-I), Line,
                                         UnitOptions, Options, Body)
                      )),
                      Time),
@@ -1058,7 +1071,7 @@ inherit_option(Name, Options0, Chain, Default, Options) :-
 	Options = [Opt|Options0]
     ).
 
-%!  run_test_once(+Unit, +Name, Progress, +Line, +Options, +Body)
+%!  run_test_once(+Unit, +Name, +Progress, +Line, +Options, +Body)
 %
 %   Deal with occurs_check, i.e., running the  test multiple times with different
 %   unification settings wrt. the occurs check.
@@ -1086,15 +1099,15 @@ run_test_once(Unit, Name, Progress, Line, Options, Body) :-
 
 :- det(report_result/4).
 report_result(failure(Unit, Name, Line, How, Time),
-	      Progress, Output, Options) :-
-    !,
+	      Progress, Output, Options) =>
     failure(Unit, Name, Progress, Line, How, Time, Output, Options).
 report_result(success(Unit, Name, Line, Determinism, Time),
-	      Progress, Output, Options) :-
-    !,
+	      Progress, Output, Options) =>
     success(Unit, Name, Progress, Line, Determinism, Time, Output, Options).
-report_result(setup_failed(_Unit, _Name, _Line),
-	      _Progress, _Output, _Options).
+report_result(setup_failed(Unit, Name, Line, Time, Output, Result),
+	      Progress, _Output, Options) =>
+    failure(Unit, Name, Progress, Line,
+            setup_failed(Result), Time, Output, Options).
 
 %!  run_test_6(+Unit, +Name, +Line, +Options, :Body, -Result) is det.
 %
@@ -1117,13 +1130,16 @@ report_result(setup_failed(_Unit, _Name, _Line),
 %     - setup_failed(Unit, Name, Line)
 
 run_test_6(Unit, Name, Line, Options, Body, Result) :-
-    option(setup(_Setup), Options),
+    option(setup(Setup), Options),
     !,
-    (   unit_module(Unit, Module),
-        setup(Module, test(Unit,Name,Line), Options)
+    unit_module(Unit, Module),
+    capture_output(call_time(reify(call_ex(Module, Setup), SetupResult),
+                             Time),
+                   Output),
+    (   SetupResult == true
     ->  run_test_7(Unit, Name, Line, Options, Body, Result),
         cleanup(Module, Options)
-    ;   Result = setup_failed(Unit, Name, Line)
+    ;   Result = setup_failed(Unit, Name, Line, Time, Output, SetupResult)
     ).
 run_test_6(Unit, Name, Line, Options, Body, Result) :-
     unit_module(Unit, Module),
@@ -1810,7 +1826,7 @@ progress(UnitTest, _Progress, forall(end, Nth, FTotal), Time) =>
     ),
     test_count(Total),
     job_feedback(information, progress(UnitTest, forall(FTotal,FFailed), Nth/Total, Time)).
-progress(UnitTest, Progress, Result, Time), Progress = forall(_Vars, Nth-_I) =>
+progress(UnitTest, Progress, Result, Time), Progress = forall(_Gen, Nth-_I) =>
     with_mutex(plunit_forall_counter,
                update_forall_failures(Nth, Result)),
     test_count(Total),
@@ -2000,7 +2016,7 @@ message(plunit(begin(Unit:Test, _Location, Progress))) -->
 message(plunit(end(_UnitTest, _Location, _Progress))) -->
     [].
 message(plunit(progress(_UnitTest, Status, _Progress, _Time))) -->
-    { Status = forall(_,_)
+    { Status = forall(_Gen,_NthI)
     ; Status == assertion
     },
     !.
@@ -2125,11 +2141,11 @@ message(concurrent) -->
       'See set_test_options/1 with jobs(Count) for concurrent testing.'
     ].
 
-test_name(Name, forall(Bindings, _Nth-I)) -->
+test_name(Name, forall(Generator, _Nth-I)/_Total) -->
     !,
     test_name(Name, -),
-    [ ' (~d-th forall bindings = '-[I],
-      ansi(code, '~p', [Bindings]), ')'-[]
+    [ ' (~d-th forall generator = '-[I],
+      ansi(code, '~p', [Generator]), ')'-[]
     ].
 test_name(Name, _) -->
     !,
@@ -2233,6 +2249,15 @@ failure(throw(Error)) -->
 failure(message) -->
     !,
     [ 'Generated unexpected warning or error'-[] ].
+failure(setup_failed(throw(Error))) -->
+    { Error = error(_,_),
+      !,
+      message_to_string(Error, Message)
+    },
+    [ 'test setup goal raised error: ~w'-[Message] ].
+failure(setup_failed(_)) -->
+    !,
+    [ 'test setup goal failed' ].
 failure(Why) -->
     [ '~p'-[Why] ].
 
