@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2021, University of Amsterdam
+    Copyright (c)  1985-2024, University of Amsterdam
 			      CWI, Amsterdam
 			      SWI-Prolog Solutions b.v.
     All rights reserved.
@@ -79,7 +79,7 @@ typedef struct _operator		/* storage in tables */
 
 typedef struct _opdef			/* predefined and enumerated */
 { atom_t name;
-  short  type;
+  unsigned char type;
   short  priority;
 } opdef;
 
@@ -90,27 +90,28 @@ typedef struct _opdef			/* predefined and enumerated */
 		 *******************************/
 
 static void
-copyOperatorSymbol(void *name, void **value)
-{ operator *op = *value;
+copyOperatorSymbol(table_key_t name, table_value_t *value)
+{ operator *op = val2ptr(*value);
   operator *o2 = allocHeapOrHalt(sizeof(*o2));
 
+  PL_register_atom((atom_t) name);
   *o2 = *op;
-  *value = o2;
+  *value = ptr2val(o2);
 }
 
 
 static void
-freeOperatorSymbol(void *name, void *value)
-{ operator *op = value;
+freeOperatorSymbol(table_key_t name, table_value_t value)
+{ operator *op = val2ptr(value);
 
   PL_unregister_atom((atom_t) name);
   freeHeap(op, sizeof(*op));
 }
 
 
-static Table
+static TableWP
 newOperatorTable(int size)
-{ Table t = newHTable(size);
+{ TableWP t = newHTableWP(size);
 
   t->copy_symbol = copyOperatorSymbol;
   t->free_symbol = freeOperatorSymbol;
@@ -129,11 +130,11 @@ tables.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
-defOperator(Module m, atom_t name, int type, int priority, int force)
+defOperator(Module m, atom_t name, unsigned char type, short priority, int force)
 { GET_LD
   operator *op;
   int t = (type & OP_MASK);		/* OP_PREFIX, ... */
-  int must_reg = FALSE;
+  int must_reg = false;
 
   DEBUG(MSG_OPERATOR, Sdprintf(":- op(%d, %s, %s) in module %s\n",
 			       priority,
@@ -162,11 +163,11 @@ defOperator(Module m, atom_t name, int type, int priority, int force)
   if ( !m->operators )
     m->operators = newOperatorTable(8);
 
-  if ( (op = lookupHTable(m->operators, (void *)name)) )
+  if ( (op = lookupHTableWP(m->operators, name)) )
   { ;
   } else if ( priority < 0 )
   { PL_UNLOCK(L_OP);				/* already inherited: do not change */
-    return TRUE;
+    return true;
   } else
   { op = allocHeapOrHalt(sizeof(*op));
 
@@ -177,18 +178,18 @@ defOperator(Module m, atom_t name, int type, int priority, int force)
     op->type[OP_INFIX]       = OP_INHERIT;
     op->type[OP_POSTFIX]     = OP_INHERIT;
 
-    must_reg = TRUE;
+    must_reg = true;
   }
 
   op->priority[t] = priority;
   op->type[t]     = (priority >= 0 ? type : OP_INHERIT);
   if ( must_reg )
   { PL_register_atom(name);
-    addNewHTable(m->operators, (void *)name, op);
+    addNewHTableWP(m->operators, name, op);
   }
   PL_UNLOCK(L_OP);
 
-  return TRUE;
+  return true;
 }
 
 
@@ -208,7 +209,7 @@ visibleOperator(Module m, atom_t name, int kind)
   ListCell c;
 
   if ( m->operators &&
-       (op = lookupHTable(m->operators, (void *)name)) )
+       (op = lookupHTableWP(m->operators, name)) )
   { if ( op->type[kind] != OP_INHERIT )
       return op;
   }
@@ -222,7 +223,7 @@ visibleOperator(Module m, atom_t name, int kind)
 
 
 int
-currentOperator(Module m, atom_t name, int kind, int *type, int *priority)
+currentOperator(Module m, atom_t name, int kind, unsigned char *type, short *priority)
 { operator *op;
 
   assert(kind >= OP_PREFIX && kind <= OP_POSTFIX);
@@ -278,7 +279,8 @@ scanPriorityOperator(Module m, atom_t name, int *done, int sofar)
   if ( *done != 0x7 )
   { operator *op;
 
-    if ( m->operators && (op = lookupHTable(m->operators, (void *)name)) )
+    if ( m->operators &&
+	 (op = lookupHTableWP(m->operators, name)) )
       sofar = maxOp(op, done, sofar);
 
     if ( *done != 0x7 )
@@ -310,7 +312,7 @@ priorityOperator(Module m, atom_t name)
 		 *	  PROLOG BINDING	*
 		 *******************************/
 
-static int
+static unsigned char
 atomToOperatorType(atom_t atom)
 { if (atom == ATOM_fx)			return OP_FX;
   else if (atom == ATOM_fy)		return OP_FY;
@@ -346,7 +348,7 @@ PRED_IMPL("op", 3, op, PL_FA_TRANSPARENT|PL_FA_ISO)
 { PRED_LD
   atom_t nm;
   atom_t tp;
-  int t;
+  unsigned char t;
   int p;
   Module m = MODULE_parse;
 
@@ -355,7 +357,7 @@ PRED_IMPL("op", 3, op, PL_FA_TRANSPARENT|PL_FA_ISO)
   term_t name = A3;
 
   if ( !PL_strip_module(name, &m, name) )
-    return FALSE;
+    return false;
   if ( m == MODULE_system )
   { term_t t = PL_new_term_ref();
     term_t a = PL_new_term_ref();
@@ -368,31 +370,31 @@ PRED_IMPL("op", 3, op, PL_FA_TRANSPARENT|PL_FA_ISO)
   }
 
   if ( !PL_get_atom_ex(type, &tp) )
-    fail;
+    return false;
   if ( !PL_get_integer_ex(pri, &p) )
-    fail;
+    return false;
   if ( !((p >= 0 && p <= OP_MAXPRIORITY) || (p == -1 && m != MODULE_user)) )
     return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_operator_priority, pri);
   if ( !(t = atomToOperatorType(tp)) )
     return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_operator_specifier, type);
 
   if ( PL_get_atom(name, &nm) )
-  { return defOperator(m, nm, t, p, FALSE);
+  { return defOperator(m, nm, t, (short)p, false);
   } else
   { term_t l = PL_copy_term_ref(name);
     term_t e = PL_new_term_ref();
 
-    while( PL_get_list(l, e, l) )
+    while( PL_get_list_ex(l, e, l) )
     { if ( !PL_get_atom(e, &nm) )
 	return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_atom, e);
-      if ( !defOperator(m, nm, t, p, FALSE) )
-	return FALSE;
+      if ( !defOperator(m, nm, t, (short)p, false) )
+	return false;
     }
-    if ( !PL_get_nil(l) )
-      return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_list, l);
+    if ( !PL_get_nil_ex(l) )
+      return false;
   }
 
-  succeed;
+  return true;
 }
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -402,7 +404,7 @@ of matching operators and (on backtracking) return the matching ones.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static void
-addOpToBuffer(Buffer b, atom_t name, int type, int priority)
+addOpToBuffer(Buffer b, atom_t name, unsigned char type, short priority)
 { opdef *op = baseBuffer(b, opdef);
   int mx    = (int)entriesBuffer(b, opdef);
   int i;
@@ -421,13 +423,16 @@ addOpToBuffer(Buffer b, atom_t name, int type, int priority)
 
 
 static void
-addOpsFromTable(Table t, atom_t name, int priority, int type, Buffer b)
-{ TableEnum e = newTableEnum(t);
-  atom_t nm;
-  operator *op;
+addOpsFromTable(TableWP t, atom_t name, int priority, int type, Buffer b)
+{ TableEnum e = newTableEnumWP(t);
+  table_key_t tk;
+  table_value_t tv;
 
-  while( advanceTableEnum(e, (void**)&nm, (void**)&op) )
-  { if ( nm == name || name == NULL_ATOM )
+  while( advanceTableEnum(e, &tk, &tv) )
+  { atom_t nm = (atom_t)tk;
+    operator *op = val2ptr(tv);
+
+    if ( nm == name || name == NULL_ATOM )
     { if ( type )
       { int kind = type&OP_MASK;
 
@@ -482,10 +487,11 @@ typedef struct
 } op_enum;
 
 
-static word
-current_op(Module m, int inherit,
+#define current_op(m, inherit, prec, type, name, h) LDFUNC(current_op, m, inherit, prec, type, name, h)
+static foreign_t
+current_op(DECL_LD Module m, int inherit,
 	   term_t prec, term_t type, term_t name,
-	   control_t h ARG_LD)
+	   control_t h)
 { op_enum *e;
   Buffer b;
   int mx;
@@ -501,7 +507,7 @@ current_op(Module m, int inherit,
       if ( PL_is_variable(name) )
 	nm = NULL_ATOM;
       else if ( !PL_get_atom_ex(name, &nm) )
-	return FALSE;
+	return false;
 
       if ( PL_is_variable(prec) )
       { p = 0;
@@ -509,7 +515,7 @@ current_op(Module m, int inherit,
       { if ( !(p > 0 && p <= 1200) )
 	  return PL_domain_error("operator_priority", prec);
       } else
-      { return FALSE;
+      { return false;
       }
 
       if ( PL_is_variable(type) )
@@ -519,7 +525,7 @@ current_op(Module m, int inherit,
 	  return PL_error(NULL, 0, NULL, ERR_DOMAIN,
 			  ATOM_operator_specifier, type);
       } else
-	return FALSE;
+	return false;
 
       e = allocHeapOrHalt(sizeof(*e));
       b = &e->buffer;
@@ -567,7 +573,7 @@ current_op(Module m, int inherit,
     { if ( e->index == mx )
       { discardBuffer(&e->buffer);
         freeHeap(e, sizeof(*e));
-	return TRUE;
+	return true;
       }
       ForeignRedoPtr(e);
     }
@@ -581,26 +587,27 @@ current_op(Module m, int inherit,
   fail;
 }
 
+#define get_op_module(a3, name, m) LDFUNC(get_op_module, a3, name, m)
 static int
-get_op_module(term_t a3, term_t name, Module *m ARG_LD)
+get_op_module(DECL_LD term_t a3, term_t name, Module *m)
 { atom_t mname = 0;
   Word p;
   word w;
 
-  if ( !(p=stripModuleName(valTermRef(a3), &mname PASS_LD)) ||
+  if ( !(p=stripModuleName(valTermRef(a3), &mname)) ||
        !(w=linkValG(p)) )
-    return FALSE;
+    return false;
   *valTermRef(name) = w;
 
   if ( *m && (*m)->name == mname )
-    return TRUE;
+    return true;
   if ( mname && !(*m=acquireModule(mname)) )
   { DEBUG(MSG_OPERATOR, Sdprintf("Could not acquire module %s\n",
 				 PL_atom_chars(mname)));
     *m = MODULE_user;
   }
 
-  return TRUE;
+  return true;
 }
 
 
@@ -611,15 +618,15 @@ PRED_IMPL("current_op", 3, current_op,
   Module mp = MODULE_parse;
   Module m  = mp;
   term_t name = A3;
-  word rc;
+  foreign_t rc;
 
   if ( CTX_CNTRL != FRG_CUTTED )
   { if ( !(name = PL_new_term_ref()) ||
-	 !get_op_module(A3, name, &m PASS_LD) )
-      return FALSE;
+	 !get_op_module(A3, name, &m) )
+      return false;
   }
 
-  rc = current_op(m, TRUE, A1, A2, name, PL__ctx PASS_LD);
+  rc = current_op(m, true, A1, A2, name, PL__ctx);
   if ( m != mp )
     releaseModule(m);
 
@@ -642,11 +649,11 @@ PRED_IMPL("$local_op", 3, local_op, PL_FA_NONDETERMINISTIC|PL_FA_TRANSPARENT)
 
   if ( CTX_CNTRL != FRG_CUTTED )
   { if ( !(name = PL_new_term_ref()) ||
-	 !get_op_module(A3, name, &m PASS_LD) )
-      return FALSE;
+	 !get_op_module(A3, name, &m) )
+      return false;
   }
 
-  return current_op(m, FALSE, A1, A2, name, PL__ctx PASS_LD);
+  return current_op(m, false, A1, A2, name, PL__ctx);
 }
 
 
@@ -730,14 +737,14 @@ void
 initOperators(void)
 { const opdef *op;
 
-  MODULE_system->operators = newOperatorTable(32);
+  MODULE_system->operators = newOperatorTable(128);
 
   for( op = operators; op->name; op++ )
-    defOperator(MODULE_system, op->name, op->type, op->priority, TRUE);
+    defOperator(MODULE_system, op->name, op->type, op->priority, true);
 
   if ( !GD->options.traditional )
-  { defOperator(MODULE_system, PL_new_atom("."), OP_YFX, 100, TRUE);
-    defOperator(MODULE_system, ATOM_colon_eq,    OP_XFX, 800, TRUE);
+  { defOperator(MODULE_system, PL_new_atom("."), OP_YFX, 100, true);
+    defOperator(MODULE_system, ATOM_colon_eq,    OP_XFX, 800, true);
   }
 }
 
